
-- Requires libsocket

-- TODO:
-- do pubsub tunes+notify instead of status hacking

require 'socket'

local settings = {
	hostname     = "localhost",
	password     = nil,
	port         = 6600,
}

mpd = {}

-- separator allows split output into records, that are started by any of present in separator keys
-- returns table of field (lowercase) - value records
-- command status is returned in STATUS field
-- on unexpected errors returns just false, dropping any available data
function mpd.receive_message ( tcp, separator )
	local ret  = {}
	local line = tcp:receive ( '*l' )
	while line and not ( line:find ( '^OK' ) or line:find ( '^ACK' ) ) do
		local key, val = line:match ( '^(.-):%s*(.*)$' )
		if key then
			if separator then
				key = key:lower ()
				if separator[key] then
					table.insert ( ret, {} )
				end
				ret[#ret][key]   = val
			else
				ret[key:lower()] = val
			end
		end
		line = tcp:receive ( '*l' )
	end
	if not line then
		return false -- an error occured, ret, most likely, does not contains exact and complete info...
	else
		ret.STATUS = line:match ( '^%S+' )
		return ret
	end
end

-- use: mpd.call_command { 'status' }
--      mpd.call_command { 'lsinfo misc', list = { file = true, directory = true } }
--      mpd.call_command { 'next', noret = true }
--      mpd.call_command { 'status', 'currentsong' }
-- on one command returns just results of that command, on multi - array of results
-- on errors returns nil
-- noret can terminate socket too early, thus, do not use it with lists of commands
function mpd.call_command ( opts )
	local tcp = socket.tcp ()
	if not tcp then
		print ( 'mpd: cannot get master tcp object' )
		return nil
	elseif not tcp:connect ( settings.hostname, settings.port ) then
		tcp:close ()
		print ( 'mpd: cannot connect to server' )
		return nil
	end

	local ret = {}
	if not opts.noret then
		ret = mpd.receive_message ( tcp )
		if not ret then
			tcp:close ()
			print ( 'mpd: error getting greeting from server' )
			return nil
		elseif ret.STATUS ~= 'OK' then
			print ( 'mpd: server ack\'s in greeting' )
		end
	end

	for num, command in ipairs ( opts ) do
		if not tcp:send ( command .. "\n" ) then
			tcp:close ()
			print ( 'mpd: error sending command ' .. command )
			return nil
		end
		if not opts.noret then
			ret[num] = mpd.receive_message ( tcp, opts.list )
			if not ret[num] then
				print ( 'mpd: error getting result' )
			end
			if ret[num]['STATUS'] ~= 'OK' then
				print ( 'mpd: server acks our command ' .. command )
			end
		end
	end

	tcp:close ()
	if #ret > 1 then
		return ret
	else
		return ret[1]
	end
end

-- MCABBER PART --

mpd_enabled = false

-- XXX: to C?
char2status = {
	f = 'free',
	o = 'online',
	a = 'away',
	d = 'dnd',
	n = 'notavail',
	i = 'invisible',
	['_'] = 'offline',
	['?'] = 'message',
}

mpd_pub_song = { xmlns = 'http://jabber.org/protocol/tune' }

function mpd_getstatus ()
	if not mpd_enabled then
		mpd_pub_song.artist = nil
		mpd_pub_song.length = nil
		mpd_pub_song.source = nil
		mpd_pub_song.title  = nil
		mpd_pub_song.track  = nil
		return mpd_pub_song
	end
	
	local stats = mpd.call_command { 'status', 'currentsong' }
	if stats[1].state ~= 'play' and stats[1].state ~= 'pause' then
		mpd_pub_song.artist = nil
		mpd_pub_song.length = nil
		mpd_pub_song.source = nil
		mpd_pub_song.title  = nil
		mpd_pub_song.track  = nil
		return mpd_pub_song
	end

	local dir, file = stats[2].file:match ( '(.+)/(.-)' )
	local modified  = false
	-- populate according to currentsong fields: artist - artist, length - time, source - album, title - title, track - id, rating - ?, uri - ?
	local artist, length, source, title, track = stats[2].artist, stats[2].time, stats[2].album, stats[2].title, stats[2].id
	
	if not artist or artist == '' then
		artist = 'Unknown'
	end
	if not mpd_pub_song.artist or artist ~= mpd_pub_song.artist[1] then
		mpd_pub_song.artist = { artist }
		modified = true
	end

	if length and length ~= 0 then
		if not mpd_pub_song.length or length ~= mpd_pub_song.length[1] then
			mpd_pub_song.length = { length }
			modified = true
		end
	elseif mpd_pub_song.length then -- no length
		mpd_pub_song.length = nil
		modified = true
	end

	if not source or source == '' then
		source = dir
	end
	if not mpd_pub_song.source or source ~= mpd_pub_song.source[1] then
		mpd_pub_song.source = { source }
		modified = true
	end

	if not title or title == '' then
		title = file
	end
	if not mpd_pub_song.title or title ~= mpd_pub_song.title[1] then
		mpd_pub_song.title = { title }
		modified = true
	end

	if not mpd_pub_song.track or track ~= mpd_pub_song.track[1] then
		mpd_pub_song.track = { track }
		modified = true
	end

	if modified then
		return mpd_pub_song
	else
		return nil
	end
end

function parse_status ()
	local stletter, stmessage = main.status ()
	local cmd = char2status[stletter]
	local message, mpd_string = stmessage:match ( "^(.-)%s+(%[mpd:%s+.+%s*%])" )
	if message then
		return cmd, message, mpd_string
	else
		return cmd, stmessage, ''
	end
end

function mpd_callback ()
	local sdata = mpd_getstatus ()
	if sdata then
		local conn = lm.connection.bless ( main.connection () )
		if conn:status () == 'authenticated' then
			conn:send (
				lm.message.create { mtype = 'iq-set', from = conn:jid (),
					pubsub = { xmlns = 'http://jabber.org/protocol/pubsub',
						publish = { node = 'http://jabber.org/protocol/tune',
							item = {
								tune = sdata,
							},
						},
					},
				})
		end
	end
	if mpd_enabled then
		return true
	else
		return false
	end
end

-- do not call it too fast, or you end up with many daemons at once
function enable_mpd ( yn )
	if yn == nil then
		yn = true
	end
	if yn then
		if not mpd_enabled then
			main.timer ( 15, mpd_callback )
			mpd_enabled = true
			-- update status
		end
	else
		if mpd_enabled then
			mpd_enabled = false
			-- update status
		end
	end
end

main.command ( 'mpd',
	function ( args )
		local enable = yesno ( args )
		if enable == nil then
			if mpd_enabled then
				print ( "MPD status string is enabled" )
			else
				print ( "MPD status string is disabled" )
			end
		else
			enable_mpd ( enable )
		end
	end, boolean_cid )

commands_help['mpd'] = "[enable|disable|on|off|yes|no|true|false]\n\nSets state of mpd string in your status.\nIf state is omitted, prints current state."

mpd_incoming_message_handler = lm.message_handler.new (
	function ( conn, mess )
		-- we can add that validation stuff later, if it will be necessary
		local tune = mess:path ( 'event', 'items', 'item', 'tune' )
		if tune then
			local from = mess:attribute ( 'from' )
			local item = tune:children ()
			main.print_info ( from, "Tunes notification:" )
			while item do
				main.print_info ( from, ("- %s: %s"):format ( item:name (), item:value () ) )
				item = item:next ()
			end
			return true
		end
		return false
	end )

mpd_handler_registered = false

hooks_d['hook-post-connect'].mpd =
	function ( args )
		lm.connection.bless( main.connection () ):handler ( mpd_incoming_message_handler, 'message', 'normal' )
		mpd_handler_registered = true
		hooks_d['hook-post-connect'].mpd =
			function ( args )
				if mpd_enabled then
					mpd_callback ()
				end
			end
		hooks_d['hook-quit'].mpd =
			function ( args )
				if mpd_handler_registered then
					lm.connection.bless( main.connection () ):handler ( mpd_incoming_message_handler, 'message' )
					mpd_handler_registered = false
				end
			end
		if mpd_enabled then
			mpd_callback ()
		end
	end

-- XXX: this really should be initialized after connection establishment?
-- but as this thing is implemented by now, it will be cached by server,
-- and, thus, we will be unable to get notifications.
main.add_feature ( 'http://jabber.org/protocol/tune' )
main.add_feature ( 'http://jabber.org/protocol/tune+notify' )

-- vim: se ts=4: --
